@@ -4,6 +4,7 @@ import (
 	"crypto/tls"
 	"fmt"
 	"log"
+	"net"
 	"net/http"
 	"os"
 
@@ -14,8 +15,6 @@ import (
 	"github.com/quic-go/quic-go/interop/utils"
 )
 
-var tlsConf *tls.Config
-
 func main() {
 	logFile, err := os.Create("/logs/log.txt")
 	if err != nil {
@@ -45,20 +44,22 @@ func main() {
 		fmt.Println(err)
 		os.Exit(1)
 	}
-	tlsConf = &tls.Config{
+	tlsConf := &tls.Config{
 		Certificates: []tls.Certificate{cert},
 		KeyLogWriter: keyLog,
+		NextProtos:   []string{http09.NextProto},
 	}
 
 	switch testcase {
 	case "versionnegotiation", "handshake", "retry", "transfer", "resumption", "multiconnect", "zerortt":
-		err = runHTTP09Server(quicConf, testcase == "retry")
+		err = runHTTP09Server(tlsConf, quicConf, testcase == "retry")
 	case "chacha20":
 		reset := qtls.SetCipherSuite(tls.TLS_CHACHA20_POLY1305_SHA256)
 		defer reset()
-		err = runHTTP09Server(quicConf, false)
+		err = runHTTP09Server(tlsConf, quicConf, false)
 	case "http3":
-		err = runHTTP3Server(quicConf)
+		tlsConf.NextProtos = []string{http3.NextProtoH3}
+		err = runHTTP3Server(tlsConf, quicConf)
 	default:
 		fmt.Printf("unsupported test case: %s\n", testcase)
 		os.Exit(127)
@@ -70,20 +71,30 @@ func main() {
 	}
 }
 
-func runHTTP09Server(quicConf *quic.Config, forceRetry bool) error {
-	server := http09.Server{
-		Server: &http.Server{
-			Addr:      ":443",
-			TLSConfig: tlsConf,
-		},
-		ForceRetry: forceRetry,
-		QuicConfig: quicConf,
-	}
+func runHTTP09Server(tlsConf *tls.Config, quicConf *quic.Config, forceRetry bool) error {
 	http.DefaultServeMux.Handle("/", http.FileServer(http.Dir("/www")))
-	return server.ListenAndServe()
+	server := http09.Server{}
+
+	udpAddr, err := net.ResolveUDPAddr("udp", ":443")
+	if err != nil {
+		return err
+	}
+	conn, err := net.ListenUDP("udp", udpAddr)
+	if err != nil {
+		return err
+	}
+	tr := &quic.Transport{
+		Conn:                conn,
+		VerifySourceAddress: func(net.Addr) bool { return forceRetry },
+	}
+	ln, err := tr.ListenEarly(tlsConf, quicConf)
+	if err != nil {
+		return err
+	}
+	return server.ServeListener(ln)
 }
 
-func runHTTP3Server(quicConf *quic.Config) error {
+func runHTTP3Server(tlsConf *tls.Config, quicConf *quic.Config) error {
 	server := http3.Server{
 		Addr:       ":443",
 		TLSConfig:  tlsConf,
