@@ -23,20 +23,21 @@ import (
 const testPackerConnIDLen = 4
 
 type testPacketPacker struct {
-	packer                         *packetPacker
-	initialStream, handshakeStream *cryptoStream
-	datagramQueue                  *datagramQueue
-	pnManager                      *mockackhandler.MockSentPacketHandler
-	sealingManager                 *MockSealingManager
-	framer                         *MockFrameSource
-	ackFramer                      *MockAckFrameSource
-	retransmissionQueue            *retransmissionQueue
+	packer              *packetPacker
+	initialStream       *initialCryptoStream
+	handshakeStream     *cryptoStream
+	datagramQueue       *datagramQueue
+	pnManager           *mockackhandler.MockSentPacketHandler
+	sealingManager      *MockSealingManager
+	framer              *MockFrameSource
+	ackFramer           *MockAckFrameSource
+	retransmissionQueue *retransmissionQueue
 }
 
 func newTestPacketPacker(t *testing.T, mockCtrl *gomock.Controller, pers protocol.Perspective) *testPacketPacker {
 	destConnID := protocol.ParseConnectionID([]byte{1, 2, 3, 4})
 	require.Equal(t, testPackerConnIDLen, destConnID.Len())
-	initialStream := newCryptoStream()
+	initialStream := newInitialCryptoStream(pers == protocol.PerspectiveClient)
 	handshakeStream := newCryptoStream()
 	pnManager := mockackhandler.NewMockSentPacketHandler(mockCtrl)
 	framer := NewMockFrameSource(mockCtrl)
@@ -121,6 +122,8 @@ func expectAppendFrames(framer *MockFrameSource, controlFrames []ackhandler.Fram
 }
 
 func TestPackLongHeaders(t *testing.T) {
+	skipIfDisableScramblingEnvSet(t)
+
 	const maxPacketSize protocol.ByteCount = 1234
 	mockCtrl := gomock.NewController(t)
 	tp := newTestPacketPacker(t, mockCtrl, protocol.PerspectiveClient)
@@ -139,7 +142,8 @@ func TestPackLongHeaders(t *testing.T) {
 	tp.sealingManager.EXPECT().Get1RTTSealer().Return(nil, handshake.ErrKeysNotYetAvailable)
 	tp.ackFramer.EXPECT().GetAckFrame(protocol.EncryptionInitial, now, false)
 	// don't EXPECT any calls for a Handshake ACK frame
-	tp.initialStream.Write([]byte("initial"))
+	clientHello := getClientHello(t, "quic-go.net")
+	tp.initialStream.Write(clientHello)
 	tp.packer.retransmissionQueue.addHandshake(&wire.PingFrame{})
 
 	p, err := tp.packer.PackCoalescedPacket(false, maxPacketSize, now, protocol.Version1)
@@ -148,8 +152,10 @@ func TestPackLongHeaders(t *testing.T) {
 	require.Len(t, p.longHdrPackets, 2)
 	require.Nil(t, p.shortHdrPacket)
 	require.Equal(t, protocol.EncryptionInitial, p.longHdrPackets[0].EncryptionLevel())
-	require.Len(t, p.longHdrPackets[0].frames, 1)
-	require.Equal(t, []byte("initial"), p.longHdrPackets[0].frames[0].Frame.(*wire.CryptoFrame).Data)
+	require.Len(t, p.longHdrPackets[0].frames, 3)
+	for _, f := range p.longHdrPackets[0].frames {
+		require.IsType(t, &wire.CryptoFrame{}, f.Frame)
+	}
 	require.Equal(t, protocol.EncryptionHandshake, p.longHdrPackets[1].EncryptionLevel())
 	require.Len(t, p.longHdrPackets[1].frames, 1)
 	require.IsType(t, &wire.PingFrame{}, p.longHdrPackets[1].frames[0].Frame)
@@ -341,7 +347,7 @@ func testPackConnectionCloseCoalesced(t *testing.T, pers protocol.Perspective) {
 	}
 	p, err := tp.packer.PackApplicationClose(&qerr.ApplicationError{
 		ErrorCode:    0x1337,
-		ErrorMessage: "test error",
+		ErrorMessage: "foobar",
 	}, maxPacketSize, protocol.Version1)
 	require.NoError(t, err)
 	switch pers {
@@ -388,7 +394,7 @@ func testPackConnectionCloseCoalesced(t *testing.T, pers protocol.Perspective) {
 	}
 	require.True(t, ccf.IsApplicationError)
 	require.Equal(t, uint64(0x1337), ccf.ErrorCode)
-	require.Equal(t, "test error", ccf.ReasonPhrase)
+	require.Equal(t, "foobar", ccf.ReasonPhrase)
 
 	// the client needs to pad this packet to the max packet size
 	switch pers {
@@ -434,7 +440,7 @@ func TestPackConnectionClose1RTT(t *testing.T) {
 	// expect no framer.PopStreamFrames
 	p, err := tp.packer.PackConnectionClose(&qerr.TransportError{
 		ErrorCode:    qerr.CryptoBufferExceeded,
-		ErrorMessage: "test error",
+		ErrorMessage: "foo",
 	}, protocol.MaxByteCount, protocol.Version1)
 	require.NoError(t, err)
 	require.Empty(t, p.longHdrPackets)
@@ -443,7 +449,7 @@ func TestPackConnectionClose1RTT(t *testing.T) {
 	ccf := p.shortHdrPacket.Frames[0].Frame.(*wire.ConnectionCloseFrame)
 	require.False(t, ccf.IsApplicationError)
 	require.Equal(t, uint64(qerr.CryptoBufferExceeded), ccf.ErrorCode)
-	require.Equal(t, "test error", ccf.ReasonPhrase)
+	require.Equal(t, "foo", ccf.ReasonPhrase)
 }
 
 func TestPack1RTTPacketNothingToSend(t *testing.T) {
@@ -729,11 +735,16 @@ func TestPackLongHeaderPadToAtLeast4Bytes(t *testing.T) {
 	// first bytes should be 2 PADDING frames...
 	require.Equal(t, []byte{0, 0}, data[:2])
 	// ...followed by the PING frame
-	frameParser := wire.NewFrameParser(false)
-	l, frame, err := frameParser.ParseNext(data[2:], protocol.EncryptionHandshake, protocol.Version1)
+	frameParser := wire.NewFrameParser(false, false)
+
+	frameType, lt, err := frameParser.ParseType(data[2:], protocol.EncryptionHandshake)
+	require.NoError(t, err)
+	require.Equal(t, 1, lt)
+	frame, l, err := frameParser.ParseLessCommonFrame(frameType, data[2+lt:], protocol.Version1)
 	require.NoError(t, err)
 	require.IsType(t, &wire.PingFrame{}, frame)
-	require.Equal(t, sealer.Overhead(), len(data)-2-l)
+	require.Zero(t, l)
+	require.Equal(t, sealer.Overhead(), len(data)-2-lt)
 }
 
 func TestPackShortHeaderPadToAtLeast4Bytes(t *testing.T) {
@@ -767,15 +778,21 @@ func TestPackShortHeaderPadToAtLeast4Bytes(t *testing.T) {
 	require.Equal(t, byte(0), payload[0])
 
 	// ... followed by the STREAM frame
-	frameParser := wire.NewFrameParser(true)
-	frameLen, frame, err := frameParser.ParseNext(payload[1:], protocol.Encryption1RTT, protocol.Version1)
+	frameParser := wire.NewFrameParser(false, false)
+	frameType, l, err := frameParser.ParseType(payload[1:], protocol.Encryption1RTT)
+	require.NoError(t, err)
+	require.Equal(t, 1, l)
+	require.True(t, frameType.IsStreamFrameType())
+
+	frame, frameLen, err := wire.ParseStreamFrame(payload[1+l:], frameType, protocol.Version1)
 	require.NoError(t, err)
 	require.Equal(t, f, frame)
-	require.Equal(t, len(payload)-1, frameLen)
+	require.Equal(t, len(payload)-2, frameLen)
 }
 
 func TestPackInitialProbePacket(t *testing.T) {
 	t.Run("client", func(t *testing.T) {
+		t.Setenv(disableClientHelloScramblingEnv, "true")
 		testPackProbePacket(t, protocol.EncryptionInitial, protocol.PerspectiveClient)
 	})
 	t.Run("server", func(t *testing.T) {
@@ -797,19 +814,22 @@ func testPackProbePacket(t *testing.T, encLevel protocol.EncryptionLevel, perspe
 	mockCtrl := gomock.NewController(t)
 	tp := newTestPacketPacker(t, mockCtrl, perspective)
 
+	var cryptoData []byte
 	switch encLevel {
 	case protocol.EncryptionInitial:
 		tp.sealingManager.EXPECT().GetInitialSealer().Return(newMockShortHeaderSealer(mockCtrl), nil)
-		tp.packer.initialStream.Write([]byte("foobar"))
+		cryptoData = getClientHello(t, "")
+		tp.packer.initialStream.Write(cryptoData)
 	case protocol.EncryptionHandshake:
 		tp.sealingManager.EXPECT().GetHandshakeSealer().Return(newMockShortHeaderSealer(mockCtrl), nil)
-		tp.packer.handshakeStream.Write([]byte("foobar"))
+		cryptoData = []byte("foobar")
+		tp.packer.handshakeStream.Write(cryptoData)
 	}
 	tp.ackFramer.EXPECT().GetAckFrame(encLevel, gomock.Any(), false)
 	tp.pnManager.EXPECT().PeekPacketNumber(encLevel).Return(protocol.PacketNumber(0x42), protocol.PacketNumberLen2)
 	tp.pnManager.EXPECT().PopPacketNumber(encLevel).Return(protocol.PacketNumber(0x42))
 
-	p, err := tp.packer.MaybePackPTOProbePacket(encLevel, maxPacketSize, time.Now(), protocol.Version1)
+	p, err := tp.packer.PackPTOProbePacket(encLevel, maxPacketSize, false, time.Now(), protocol.Version1)
 	require.NoError(t, err)
 	require.NotNil(t, p)
 	require.Len(t, p.longHdrPackets, 1)
@@ -820,7 +840,7 @@ func testPackProbePacket(t *testing.T, encLevel protocol.EncryptionLevel, perspe
 		require.Equal(t, maxPacketSize, p.buffer.Len())
 	}
 	require.Len(t, packet.frames, 1)
-	require.Equal(t, &wire.CryptoFrame{Data: []byte("foobar")}, packet.frames[0].Frame)
+	require.Equal(t, cryptoData, packet.frames[0].Frame.(*wire.CryptoFrame).Data)
 	hdrs, more := parsePacket(t, p.buffer.Data)
 	require.Len(t, hdrs, 1)
 	switch encLevel {
@@ -832,17 +852,6 @@ func testPackProbePacket(t *testing.T, encLevel protocol.EncryptionLevel, perspe
 	require.Empty(t, more)
 }
 
-func TestPackProbePacketNothingToSend(t *testing.T) {
-	mockCtrl := gomock.NewController(t)
-	tp := newTestPacketPacker(t, mockCtrl, protocol.PerspectiveClient)
-	tp.sealingManager.EXPECT().GetInitialSealer().Return(newMockShortHeaderSealer(mockCtrl), nil)
-	tp.ackFramer.EXPECT().GetAckFrame(protocol.EncryptionInitial, gomock.Any(), true)
-
-	p, err := tp.packer.MaybePackPTOProbePacket(protocol.EncryptionInitial, protocol.MaxByteCount, time.Now(), protocol.Version1)
-	require.NoError(t, err)
-	require.Nil(t, p)
-}
-
 func TestPack1RTTProbePacket(t *testing.T) {
 	const maxPacketSize protocol.ByteCount = 999
 
@@ -861,7 +870,7 @@ func TestPack1RTTProbePacket(t *testing.T) {
 		},
 	)
 
-	p, err := tp.packer.MaybePackPTOProbePacket(protocol.Encryption1RTT, maxPacketSize, time.Now(), protocol.Version1)
+	p, err := tp.packer.PackPTOProbePacket(protocol.Encryption1RTT, maxPacketSize, false, time.Now(), protocol.Version1)
 	require.NoError(t, err)
 	require.NotNil(t, p)
 	require.True(t, p.IsOnlyShortHeaderPacket())
@@ -873,18 +882,61 @@ func TestPack1RTTProbePacket(t *testing.T) {
 	require.Equal(t, maxPacketSize, packet.Length)
 }
 
-func TestPackProbePacketNothingToPack(t *testing.T) {
+func TestPackPTOProbePacketNothingToPack(t *testing.T) {
+	t.Run("Initial", func(t *testing.T) {
+		testPackPTOProbePacketNothingToPack(t, protocol.EncryptionInitial)
+	})
+	t.Run("Handshake", func(t *testing.T) {
+		testPackPTOProbePacketNothingToPack(t, protocol.EncryptionHandshake)
+	})
+	t.Run("1-RTT", func(t *testing.T) {
+		testPackPTOProbePacketNothingToPack(t, protocol.Encryption1RTT)
+	})
+}
+
+func testPackPTOProbePacketNothingToPack(t *testing.T, encLevel protocol.EncryptionLevel) {
+	const maxPacketSize protocol.ByteCount = 1234
 	mockCtrl := gomock.NewController(t)
 	tp := newTestPacketPacker(t, mockCtrl, protocol.PerspectiveServer)
 
-	tp.sealingManager.EXPECT().Get1RTTSealer().Return(newMockShortHeaderSealer(mockCtrl), nil)
-	tp.pnManager.EXPECT().PeekPacketNumber(protocol.Encryption1RTT).Return(protocol.PacketNumber(0x42), protocol.PacketNumberLen2)
-	tp.ackFramer.EXPECT().GetAckFrame(protocol.Encryption1RTT, gomock.Any(), true)
-	tp.framer.EXPECT().HasData()
+	switch encLevel {
+	case protocol.EncryptionInitial:
+		tp.sealingManager.EXPECT().GetInitialSealer().Return(newMockShortHeaderSealer(mockCtrl), nil).Times(2)
+	case protocol.EncryptionHandshake:
+		tp.sealingManager.EXPECT().GetHandshakeSealer().Return(newMockShortHeaderSealer(mockCtrl), nil).Times(2)
+	case protocol.Encryption1RTT:
+		tp.sealingManager.EXPECT().Get1RTTSealer().Return(newMockShortHeaderSealer(mockCtrl), nil).Times(2)
+		tp.framer.EXPECT().HasData().Times(2)
+	}
+	tp.pnManager.EXPECT().PeekPacketNumber(encLevel).Return(protocol.PacketNumber(0x42), protocol.PacketNumberLen2).MaxTimes(2)
+	tp.ackFramer.EXPECT().GetAckFrame(encLevel, gomock.Any(), true).Times(2)
 
-	packet, err := tp.packer.MaybePackPTOProbePacket(protocol.Encryption1RTT, protocol.MaxByteCount, time.Now(), protocol.Version1)
+	// don't force a PING to be sent
+	packet, err := tp.packer.PackPTOProbePacket(encLevel, maxPacketSize, false, time.Now(), protocol.Version1)
 	require.NoError(t, err)
 	require.Nil(t, packet)
+
+	// now force a PING to be sent
+	tp.pnManager.EXPECT().PopPacketNumber(encLevel).Return(protocol.PacketNumber(0x42))
+	packet, err = tp.packer.PackPTOProbePacket(encLevel, maxPacketSize, true, time.Now(), protocol.Version1)
+	require.NoError(t, err)
+	require.NotNil(t, packet)
+	var frames []ackhandler.Frame
+	switch encLevel {
+	case protocol.EncryptionInitial, protocol.EncryptionHandshake:
+		require.Len(t, packet.longHdrPackets, 1)
+		require.Nil(t, packet.shortHdrPacket)
+		require.Equal(t, encLevel, packet.longHdrPackets[0].EncryptionLevel())
+		frames = packet.longHdrPackets[0].frames
+	case protocol.Encryption1RTT:
+		require.Empty(t, packet.longHdrPackets)
+		require.NotNil(t, packet.shortHdrPacket)
+		frames = packet.shortHdrPacket.Frames
+	}
+
+	require.Len(t, frames, 1)
+	require.Equal(t, &wire.PingFrame{}, frames[0].Frame)
+	require.Equal(t, emptyHandler{}, frames[0].Handler)
 }
 
 func TestPackMTUProbePacket(t *testing.T) {
@@ -917,14 +969,21 @@ func TestPackPathProbePacket(t *testing.T) {
 
 	p, buf, err := tp.packer.PackPathProbePacket(
 		protocol.ParseConnectionID([]byte{1, 2, 3, 4}),
-		ackhandler.Frame{Frame: &wire.PathChallengeFrame{Data: [8]byte{1, 2, 3, 4, 5, 6, 7, 8}}},
+		[]ackhandler.Frame{
+			{Frame: &wire.PathChallengeFrame{Data: [8]byte{1, 2, 3, 4, 5, 6, 7, 8}}},
+			{Frame: &wire.PathResponseFrame{Data: [8]byte{8, 7, 6, 5, 4, 3, 2, 1}}},
+		},
 		protocol.Version1,
 	)
 	require.NoError(t, err)
 	require.Equal(t, protocol.PacketNumber(0x43), p.PacketNumber)
 	require.Nil(t, p.Ack)
 	require.Empty(t, p.StreamFrames)
-	require.Equal(t, &wire.PathChallengeFrame{Data: [8]byte{1, 2, 3, 4, 5, 6, 7, 8}}, p.Frames[0].Frame)
+	require.Len(t, p.Frames, 2)
+	// the frame order is randomized
+	frames := []wire.Frame{p.Frames[0].Frame, p.Frames[1].Frame}
+	require.Contains(t, frames, &wire.PathChallengeFrame{Data: [8]byte{1, 2, 3, 4, 5, 6, 7, 8}})
+	require.Contains(t, frames, &wire.PathResponseFrame{Data: [8]byte{8, 7, 6, 5, 4, 3, 2, 1}})
 	require.Len(t, buf.Data, protocol.MinInitialPacketSize)
 	require.True(t, p.IsPathProbePacket)
 	require.False(t, p.IsPathMTUProbePacket)
